/*===========================================================================
/
/                                   ttPage
/                           eLXg\EChE
/
/============================================================================
/ Copyright (C) 1998-2018 Sota. All rights reserved.
/
/ Redistribution and use in source and binary forms, with or without
/ modification, are permitted provided that the following conditions
/ are met:
/
/  1. Redistributions of source code must retain the above copyright
/     notice, this list of conditions and the following disclaimer.
/  2. Redistributions in binary form must reproduce the above copyright
/     notice, this list of conditions and the following disclaimer in the
/     documentation and/or other materials provided with the distribution.
/
/ THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
/ IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
/ OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
/ IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
/ INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
/ BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
/ USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
/ ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
/ (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
/ THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
/============================================================================*/

#define  STRICT
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <windowsx.h>
#include <commctrl.h>
#include <tchar.h>
#include <htmlhelp.h>

#include "common.h"
#include "resource.h"


/*===== vg^Cv =====*/

static LRESULT CALLBACK ViewWndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);
static void MoveMouseCursor(HWND hWnd, int posX, int posY);
static void RePaintView(HWND hWnd);
static void ChangeDispTopLine(HWND hWnd, int num, int redraw);
static void ChangeDispTopColumn(HWND hWnd, int num, int redraw);
static void DispText(HWND hWnd, HDC hDC, int top, int btm);
static void DispOneLine(HDC hDC, int posX, int posY, LPTSTR text, int len, HDC memDC[], int tabWidth, LPTSTR rgnTop, LPTSTR rgnBtm, FILEMANTBL *curFile, int spaceWidthPx);
static int GetCharacterWidthSimple(_TCHAR text);
static int DispMarkGo(HDC hDC, int posX, int posY, _TCHAR text, int btmPosX, COLORREF textColor, COLORREF backColor, HDC memDC[]);
static int DispLineGo(HDC hDC, int posX, int posY, LPTSTR text, LPTSTR btm, int btmPosX, COLORREF textColor, COLORREF backColor);
static void InvertTextColor(HWND hWnd, HDC hDC, LPTSTR rgnTop, LPTSTR rgnBtm);
static int GetTextPixelWidth(HDC hDC, LPTSTR text, int len);
static void CheckTopAndBtmAdrs(LPTSTR *top, LPTSTR *btm);
static void CheckTopAndBtmLine(int *top, int *btm);
static void ScrollUpDownView(int num, int smooth);
static void ScrollLeftRightView(int num);
static LPTSTR AskCursorCharPos(HWND hWnd, int posX, int posY);
static void CopyToClipBoard(HWND hWnd, LPTSTR rgnTop, LPTSTR rgnBtm, LPTSTR quote);
static void AppendToClipBoard(HWND hWnd, LPTSTR rgnTop, LPTSTR rgnBtm);
static void CopyToBuffer(HWND hWnd, LPTSTR rgnTop, LPTSTR rgnBtm, LPTSTR buf, int max);
static BOOL CALLBACK FindDialogCallBack(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam);
static void SetStrToFindHist(LPTSTR str);
static void DispMouseCaret(HWND hWnd, int posX, int posY);
static void EraseMouseCaret(HWND hWnd);

/*===== [Jȃ[N =====*/

static const _TCHAR ViewClassStr[] = _T("ttPageViewClass");
static HWND hWndView = NULL;

static HBRUSH hBrushBack = NULL;

static int WinLines;
static int WinColumns;

static int CharCellHeight;
static int CharHeight;
static int CharWidth;

static int CellSep[ONELINE_MAXLEN];

static int WaitTripleClick = NO;

static int RevFindStr = YES;

static LPTSTR CaretPos = NULL;
static RECT CaretRect;

/*===== OQ =====*/

extern HFONT hFontttPage;
extern HWND hWndHelp;

GLOBALSET   GlobalSet;          /* ݒl */

int CharCellWidth;


/*-------------------------------------------------------------------------------
Name    :   MakeViewWindow
Desc    :   r[EChE쐬
Param   :   hWnd    eEChẼEChEnh
Return  :   Xe[^X (SUCCESS/FAIL)
Note    :
-------------------------------------------------------------------------------*/
int MakeViewWindow(HWND hWnd)
{
    WNDCLASSEX  wClass;
    RECT        rect;
    int         sts;

    sts = FAIL;
    wClass.cbSize        = sizeof(WNDCLASSEX);
    wClass.style         = CS_DBLCLKS;
    wClass.lpfnWndProc   = ViewWndProc;
    wClass.cbClsExtra    = 0;
    wClass.cbWndExtra    = 0;
    wClass.hInstance     = GetProgInst();
    wClass.hIcon         = NULL;
    wClass.hCursor       = LoadCursor(NULL, IDC_ARROW);
    wClass.hbrBackground = NULL;
    wClass.lpszMenuName  = NULL;
    wClass.lpszClassName = ViewClassStr;
    wClass.hIconSm       = NULL;
    RegisterClassEx(&wClass);

    GetClientRect(hWnd, &rect);

    hWndView = CreateWindow(ViewClassStr, _T(""),
                WS_CHILD | WS_BORDER | WS_HSCROLL | WS_VSCROLL,
                AskNumWinWidth(), GetToolBarHeight(), rect.right - AskNumWinWidth(), rect.bottom - GetStatusWinHeight() - GetToolBarHeight(),
                hWnd, (HMENU)1500, GetProgInst(), NULL);

    if(hWndView != NULL)
    {
        ShowWindow(hWndView, SW_SHOW);
        DragAcceptFiles(hWndView, TRUE);
        sts = SUCCESS;
    }
    return(sts);
}


/*-------------------------------------------------------------------------------
Name    :   DeleteViewWindow
Desc    :   r[EChE폜
Param   :   none
Return  :   none
Note    :
-------------------------------------------------------------------------------*/
void DeleteViewWindow(void)
{
    UnregisterClass(ViewClassStr, GetProgInst());
    return;
}


/*-------------------------------------------------------------------------------
Name    :   GetViewWinHwnd
Desc    :   r[EChẼEChEnhԂ
Param   :   none
Return  :   EChEnh
Note    :
-------------------------------------------------------------------------------*/
HWND GetViewWinHwnd(void)
{
    return(hWndView);
}


/*-------------------------------------------------------------------------------
Name    :   SetViewBackColor
Desc    :   r[EChE̔wiFݒ
Param   :   color       wiF
Return  :   none
Note    :
-------------------------------------------------------------------------------*/
void SetViewBackColor(COLORREF color)
{
    if(hBrushBack != NULL)
    {
        DeleteObject(hBrushBack);
    }
    hBrushBack = CreateSolidBrush(color);
    return;
}


/*-------------------------------------------------------------------------------
Name    :   AskViewBackColor
Desc    :   r[EChE̔wiFԂ
Param   :   none
Return  :   wiF̃uV
Note    :
-------------------------------------------------------------------------------*/
HBRUSH AskViewBackColor(void)
{
    return(hBrushBack);
}


/*-------------------------------------------------------------------------------
Name    :   ResizeViewWindow
Desc    :   r[EChẼTCYύX̏
Param   :   none
Return  :   none
Note    :
-------------------------------------------------------------------------------*/
void ResizeViewWindow(void)
{
    RECT    rect;

    GetClientRect(GetMainHwnd(), &rect);
    SetWindowPos(hWndView, NULL,
        AskNumWinWidth(),
        GetToolBarHeight(),
        rect.right - AskNumWinWidth(),
        rect.bottom - GetStatusWinHeight() - GetToolBarHeight(),
        SWP_NOACTIVATE | SWP_NOZORDER);

    if(AskCurFileTbl() != NULL)
    {
        if(AskCurExtSet()->FoldRight == YES)
        {
            MakeLineTopTbl(AskCurFileTbl());
            InvalidateRgn(hWndView, NULL, TRUE);
            InvalidateRgn(GetNumWinHwnd(), NULL, TRUE);
            DispCurrentLineNum(STATUS_WIN_TOP_LINE_NUM, AskCurFileTbl()->DispTopLine);
            DispMousePos();
        }
    }

    return;
}


/*-------------------------------------------------------------------------------
Name    :   ViewWndProc
Desc    :   r[EChẼbZ[W
Param   :   hWnd    EChEnh
            message bZ[Wԍ
            wParam  bZ[W WPARAM 
            lParam  bZ[W LPARAM 
Return  :   LRESULT
Note    :
-------------------------------------------------------------------------------*/
static LRESULT CALLBACK ViewWndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    int         tmp;
    int         posX;
    int         posY;
    POINT       csrPos;
    _TCHAR      filename[MAX_PATH+1];
    EXTSET      *curExt;
    FILEMANTBL  *curFile;
    LPTSTR      tmpPos;
    int         dist;

    curExt = AskCurExtSet();
    curFile = AskCurFileTbl();

    switch (message)
    {
        case WM_COMMAND :
            EraseMouseCaret(hWnd);
            switch(LOWORD(wParam))
            {
                case MENU_NEXTLINE :
                case MENU_PREVLINE :
                case MENU_NEXT5LINE :
                case MENU_PREV5LINE :
                case MENU_NEXT_N_LINE :
                case MENU_PREV_N_LINE :
                case MENU_NEXTPAGE :
                case MENU_PREVPAGE :
                case MENU_FILEBTM :
                case MENU_FILETOP :
                case MENU_RIGHT :
                case MENU_LEFT :
                case MENU_RIGHT5 :
                case MENU_LEFT5 :
                case MENU_FIND :
                case MENU_FINDBACKWD :
                case MENU_FINDNEXT :
                case MENU_FINDPREV :
                case MENU_FINDDISP :
                case MENU_SELALL :
                case MENU_COPY :
                case MENU_APPENDCOPY :
                case MENU_QUOTECOPY :
                case MENU_GOTO :
                    if(curFile != NULL)
                    {
                        switch(LOWORD(wParam))
                        {
                            case MENU_NEXTLINE :
                                if(curFile->DispTopLine < (curFile->TextLines - (WinLines - 1)))
                                {
                                    ChangeDispTopLine(hWnd, curFile->DispTopLine + 1, NO);
                                }
                                break;

                            case MENU_PREVLINE :
                                if(curFile->DispTopLine > 0)
                                {
                                    ChangeDispTopLine(hWnd, curFile->DispTopLine - 1, NO);
                                }
                                break;

                            case MENU_NEXT5LINE :
                                lParam = 5;
                                /* break͂Ȃ */

                            case MENU_NEXT_N_LINE :
                                if(curFile->DispTopLine < (curFile->TextLines - (WinLines - 1)))
                                {
                                    tmp = curFile->DispTopLine + lParam;
                                    if(tmp > (curFile->TextLines - (WinLines - 1)))
                                    {
                                        tmp = curFile->TextLines - (WinLines - 1);
                                    }
                                    ChangeDispTopLine(hWnd, tmp, NO);
                                }
                                break;

                            case MENU_PREV5LINE :
                                lParam = 5;
                                /* break͂Ȃ */

                            case MENU_PREV_N_LINE :
                                if(curFile->DispTopLine > 0)
                                {
                                    tmp = curFile->DispTopLine - lParam;
                                    if(tmp < 0)
                                    {
                                        tmp = 0;
                                    }
                                    ChangeDispTopLine(hWnd, tmp, NO);
                                }
                                break;

                            case MENU_NEXTPAGE :
                                if(curFile->DispTopLine < (curFile->TextLines - (WinLines - 1)))
                                {
                                    tmp = curFile->DispTopLine + (WinLines - 2);
                                    if(tmp > (curFile->TextLines - (WinLines - 1)))
                                    {
                                        tmp = curFile->TextLines - (WinLines - 1);
                                    }
                                    ChangeDispTopLine(hWnd, tmp, NO);
                                }
                                break;

                            case MENU_PREVPAGE :
                                if(curFile->DispTopLine > 0)
                                {
                                    tmp = curFile->DispTopLine - (WinLines - 2);
                                    if(tmp < 0)
                                    {
                                        tmp = 0;
                                    }
                                    ChangeDispTopLine(hWnd, tmp, NO);
                                }
                                break;

                            case MENU_FILEBTM :
                                if(curFile->DispTopLine < (curFile->TextLines - (WinLines - 1)))
                                {
                                    ChangeDispTopLine(hWnd, curFile->TextLines - (WinLines - 1), NO);
                                }
                                break;

                            case MENU_FILETOP :
                                if(curFile->DispTopLine > 0)
                                {
                                    ChangeDispTopLine(hWnd, 0, NO);
                                }
                                break;

                            case MENU_RIGHT :
                                if(curFile->DispTopColumn < (curExt->FoldPos - (WinColumns - 1)))
                                {
                                    ChangeDispTopColumn(hWnd, curFile->DispTopColumn + 1, NO);
                                }
                                break;

                            case MENU_LEFT :
                                if(curFile->DispTopColumn > 0)
                                {
                                    ChangeDispTopColumn(hWnd, curFile->DispTopColumn - 1, NO);
                                }
                                break;

                            case MENU_RIGHT5 :
                                if(curFile->DispTopColumn < (curExt->FoldPos - (WinColumns - 1)))
                                {
                                    tmp = curFile->DispTopColumn + curExt->TabWidth;
                                    if(tmp > (curExt->FoldPos - (WinColumns - 1)))
                                    {
                                        tmp = curExt->FoldPos - (WinColumns - 1);
                                    }
                                    ChangeDispTopColumn(hWnd, tmp, NO);
                                }
                                break;

                            case MENU_LEFT5 :
                                if(curFile->DispTopColumn > 0)
                                {
                                    tmp = curFile->DispTopColumn - curExt->TabWidth;
                                    if(tmp < 0)
                                    {
                                        tmp = 0;
                                    }
                                    ChangeDispTopColumn(hWnd, tmp, NO);
                                }
                                break;

                            case MENU_FIND :
                            case MENU_FINDBACKWD :
                                _tcscpy(GlobalSet.FindData.Str, _T(""));
                                if(curFile->RgnStartPos != NULL)
                                {
                                    CopyToBuffer(hWnd, curFile->RgnStartPos, curFile->RgnEndPos, GlobalSet.FindData.Str, FINDSTR_LEN+1);
                                    curFile->RgnStartPos = NULL;
                                    curFile->RgnEndPos = NULL;
                                    InvalidateRgn(hWnd, NULL, TRUE);
                                }
                                if(DialogBoxParam(GetProgInst(), MAKEINTRESOURCE(find_dlg), GetMainHwnd(), FindDialogCallBack, (LPARAM)&GlobalSet.FindData) == YES)
                                {
                                    SetStrToFindHist(GlobalSet.FindData.Str);
                                    if(LOWORD(wParam) == MENU_FIND)
                                    {
                                        dist = DIST_FORWARD;
                                    }
                                    else
                                    {
                                        dist = MENU_FINDBACKWD;
                                    }
                                    if((tmp = SearchText(curFile, curFile->DispTopLine, GlobalSet.FindData.Str, GlobalSet.FindData.CapsFlg, GlobalSet.FindData.Regexp, dist)) != -1)
                                    {
                                        RevFindStr = YES;
                                        ChangeDispTopLine(hWnd, tmp, YES);
                                    }
                                    else
                                    {
                                        InvalidateRgn(hWndView, NULL, TRUE);
                                        RingErrorBell();
                                    }
                                    MakeButtonsHide();
                                }
                                break;

                            case MENU_FINDNEXT :
                                if((tmp = SearchNextFoundLine(curFile, curFile->DispTopLine)) != -1)
                                {
                                    if(RevFindStr == YES)
                                    {
                                        ChangeDispTopLine(hWnd, tmp, NO);
                                    }
                                    else
                                    {
                                        RevFindStr = YES;
                                        ChangeDispTopLine(hWnd, tmp, YES);
                                    }
                                    MakeButtonsHide();
                                }
                                else
                                {
                                    RingErrorBell();
                                }
                                break;

                            case MENU_FINDPREV :
                                if((tmp = SearchPrevFoundLine(curFile, curFile->DispTopLine)) != -1)
                                {
                                    if(RevFindStr == YES)
                                    {
                                        ChangeDispTopLine(hWnd, tmp, NO);
                                    }
                                    else
                                    {
                                        RevFindStr = YES;
                                        ChangeDispTopLine(hWnd, tmp, YES);
                                    }
                                    MakeButtonsHide();
                                }
                                else
                                {
                                    RingErrorBell();
                                }
                                break;

                            case MENU_FINDDISP :
                                RevFindStr ^= YES;
                                InvalidateRgn(hWnd, NULL, TRUE);
                                MakeButtonsHide();
                                break;

                            case MENU_SELALL :
                                if(curFile->RgnStartPos != NULL)
                                {
                                    InvertTextColor(hWnd, NULL, curFile->RgnStartPos, curFile->RgnEndPos);
                                }
                                curFile->RgnStartPos = curFile->TextData;
                                curFile->RgnEndPos = curFile->TextData + curFile->TextSize;
                                curFile->RgnSelecting = NO;
                                MakeButtonsHide();
                                InvertTextColor(hWnd, NULL, curFile->RgnStartPos, curFile->RgnEndPos);
                                break;

                            case MENU_COPY :
                                if(curFile->RgnStartPos != NULL)
                                {
                                    CopyToClipBoard(hWnd, curFile->RgnStartPos, curFile->RgnEndPos, _T(""));
                                }
                                curFile->RgnStartPos = NULL;
                                curFile->RgnEndPos = NULL;
                                InvalidateRgn(hWnd, NULL, TRUE);
                                MakeButtonsHide();
                                break;

                            case MENU_APPENDCOPY :
                                if(curFile->RgnStartPos != NULL)
                                {
                                    AppendToClipBoard(hWnd, curFile->RgnStartPos, curFile->RgnEndPos);
                                }
                                curFile->RgnStartPos = NULL;
                                curFile->RgnEndPos = NULL;
                                InvalidateRgn(hWnd, NULL, TRUE);
                                MakeButtonsHide();
                                break;

                            case MENU_QUOTECOPY :
                                if(curFile->RgnStartPos != NULL)
                                {
                                    CopyToClipBoard(hWnd, curFile->RgnStartPos, curFile->RgnEndPos, GlobalSet.QuoteStr);
                                }
                                curFile->RgnStartPos = NULL;
                                curFile->RgnEndPos = NULL;
                                InvalidateRgn(hWnd, NULL, TRUE);
                                MakeButtonsHide();
                                break;

                            case MENU_GOTO :
                                _tcscpy(filename, _T(""));
                                if(InputDialogBox(goto_dlg, hWnd, MSGJPN_85, filename, 10, &tmp) == YES)
                                {
                                    SendMessage(GetViewWinHwnd(), WM_CHANGE_TOP, 0, _ttoi(filename)-1);
                                }
                                break;
                        }
                    }
                    break;
            }
            break;

        case WM_VSCROLL :
        case WM_HSCROLL :
        case WM_LBUTTONDOWN :
        case WM_LBUTTONUP :
        case WM_LBUTTONDBLCLK :
        case WM_LBUTTONTRICLK :
        case WM_MOUSEMOVE :
        case WM_TIMER :
            if(curFile != NULL)
            {
                switch (message)
                {
                    case WM_VSCROLL :
                        EraseMouseCaret(hWnd);
                        switch(GET_WM_VSCROLL_CODE(wParam, lParam))
                        {
                            case SB_LINEUP:
                                PostMessage(hWnd, WM_COMMAND, MAKEWPARAM(MENU_PREVLINE, 0), 0);
                                break;

                            case SB_LINEDOWN:
                                PostMessage(hWnd, WM_COMMAND, MAKEWPARAM(MENU_NEXTLINE, 0), 0);
                                break;

                            case SB_PAGEUP:
                                PostMessage(hWnd, WM_COMMAND, MAKEWPARAM(MENU_PREVPAGE, 0), 0);
                                break;

                            case SB_PAGEDOWN:
                                PostMessage(hWnd, WM_COMMAND, MAKEWPARAM(MENU_NEXTPAGE, 0), 0);
                                break;

                            case SB_THUMBTRACK:
                                    tmp = GET_WM_VSCROLL_POS(wParam, lParam) * curFile->SbarStep;
                                    if(tmp >= curFile->TextLines - 1)
                                    {
                                        tmp = curFile->TextLines - 1;
                                    }
                                    ChangeDispTopLine(hWnd, tmp, NO);
                                break;
                        }
                        break;

                    case WM_HSCROLL :
                        EraseMouseCaret(hWnd);
                        switch(GET_WM_HSCROLL_CODE(wParam, lParam))
                        {
                            case SB_LINERIGHT:
                                PostMessage(hWnd, WM_COMMAND, MAKEWPARAM(MENU_RIGHT, 0), 0);
                                break;

                            case SB_LINELEFT:
                                PostMessage(hWnd, WM_COMMAND, MAKEWPARAM(MENU_LEFT, 0), 0);
                                break;

                            case SB_PAGERIGHT:
                                PostMessage(hWnd, WM_COMMAND, MAKEWPARAM(MENU_RIGHT5, 0), 0);
                                break;

                            case SB_PAGELEFT:
                                PostMessage(hWnd, WM_COMMAND, MAKEWPARAM(MENU_LEFT5, 0), 0);
                                break;

                            case SB_THUMBTRACK:
                                    tmp = GET_WM_VSCROLL_POS(wParam, lParam);
                                    if(tmp >= curExt->FoldPos - 1)
                                    {
                                        tmp = curExt->FoldPos - 1;
                                    }
                                    ChangeDispTopColumn(hWnd, tmp, NO);
                                break;
                        }
                        break;

                    case WM_LBUTTONDOWN :
                        EraseMouseCaret(hWnd);
                        SetFocus(hWnd);
                        KillTimer(hWnd, 2);
                        if(WaitTripleClick == YES)
                        {
                            WaitTripleClick = NO;
                            PostMessage(hWnd, WM_LBUTTONTRICLK, wParam, lParam);
                        }
                        else
                        {
                            if(curFile->RgnStartPos != NULL)
                            {
                                InvertTextColor(hWnd, NULL, curFile->RgnStartPos, curFile->RgnEndPos);
                            }
                            curFile->RgnStartPos = AskCursorCharPos(hWnd, LOWORD(lParam), HIWORD(lParam));
                            curFile->RgnEndPos = curFile->RgnStartPos;
                            curFile->RgnSelecting = YES;
                            SetCapture(hWnd);
                        }
                        break;

                    case WM_LBUTTONUP :
                        EraseMouseCaret(hWnd);
                        if(curFile->RgnSelecting == YES)
                        {
                            KillTimer(hWnd, 1);
                            curFile->RgnSelecting = NO;
                            if(curFile->RgnEndPos == curFile->RgnStartPos)
                            {
                                curFile->RgnStartPos = NULL;
                                curFile->RgnEndPos = NULL;
                            }
                            ReleaseCapture();
                            MakeButtonsHide();
                        }
                        break;

                    case WM_LBUTTONDBLCLK :
                        EraseMouseCaret(hWnd);
                        if(curFile->RgnStartPos != NULL)
                        {
                            InvertTextColor(hWnd, NULL, curFile->RgnStartPos, curFile->RgnEndPos);
                        }
                        tmpPos = AskCursorCharPos(hWnd, LOWORD(lParam), HIWORD(lParam));
                        AskWordTopAndBtm(curFile, tmpPos, &curFile->RgnStartPos, &curFile->RgnEndPos);

                        curFile->RgnSelecting = NO;
                        MakeButtonsHide();
                        InvertTextColor(hWnd, NULL, curFile->RgnStartPos, curFile->RgnEndPos);
                        WaitTripleClick = YES;
                        SetTimer(hWnd, 2, GetDoubleClickTime(), NULL);
                        break;

                    case WM_LBUTTONTRICLK :
                        EraseMouseCaret(hWnd);
                        if(curFile->RgnStartPos != NULL)
                        {
                            InvertTextColor(hWnd, NULL, curFile->RgnStartPos, curFile->RgnEndPos);
                        }
                        tmpPos = AskCursorCharPos(hWnd, LOWORD(lParam), HIWORD(lParam));

                        AskPhysLineTopAndBtm(curFile, tmpPos, &curFile->RgnStartPos, &curFile->RgnEndPos);
                        curFile->RgnSelecting = NO;
                        MakeButtonsHide();
                        InvertTextColor(hWnd, NULL, curFile->RgnStartPos, curFile->RgnEndPos);
                        break;

                    case WM_MOUSEMOVE :
                        if(curFile->RgnSelecting == YES)
                        {
                            EraseMouseCaret(hWnd);
                            KillTimer(hWnd, 1);
                            posX = LOWORD(lParam);
                            posY = HIWORD(lParam);
                            if(posY > 32767)
                            {
                                posY -= 65536;
                            }
                            MoveMouseCursor(hWnd, posX, posY);
                        }
                        else
                        {
                            // }EXLbg
                            posX = LOWORD(lParam);
                            posY = HIWORD(lParam);
                            if(posY > 32767)
                            {
                                posY -= 65536;
                            }
                            DispMouseCaret(hWnd, posX, posY);
                        }
                        break;

                    case WM_TIMER :
                        EraseMouseCaret(hWnd);
                        KillTimer(hWnd, wParam);
                        if(wParam == 1)
                        {
                            GetCursorPos(&csrPos);
                            ScreenToClient(hWnd, &csrPos);
                            MoveMouseCursor(hWnd, csrPos.x, csrPos.y);
                        }
                        else if(wParam == 2)
                        {
                            WaitTripleClick = NO;
                        }
                        break;
                }
            }
            break;

        case WM_CHANGE_TOP :
            EraseMouseCaret(hWnd);
            tmp = lParam;
            if(AskCurExtSet()->LineNumType == LINENUM_WP)
            {
                if((tmp < 0) || (tmp >= curFile->TextLines))
                {
                    tmp = -1;
                }
            }
            else
            {
                tmp = GetLineNumFromLogicalLineNum(curFile, tmp);
            }

            if(tmp != -1)
            {
                ChangeDispTopLine(hWnd, tmp, NO);
            }
            else
            {
                RingErrorBell();
            }
            break;

        case WM_RBUTTONDOWN :
            EraseMouseCaret(hWnd);
            RightButtonMenu();
            break;

        case WM_DROPFILES :
            EraseMouseCaret(hWnd);
            DragQueryFile((HDROP)wParam, 0, filename, MAX_PATH);
            DragFinish((HDROP)wParam);
            OpenViewFromFile(filename, 0, ENCODE_AUTO, NO);
            MakeButtonsHide();
            break;

        case WM_PAINT :
            EraseMouseCaret(hWnd);
            RePaintView(hWnd);
            break;

        default :
            return(DefWindowProc(hWnd, message, wParam, lParam));
    }
    return(0L);
}


/*-------------------------------------------------------------------------------
Name    :   MoveMouseCursor
Desc    :   ͈͑I𒆂̃}EXJ[\̈ړ
Param   :   hWnd    EChEnh
            posX    J[\XW
            posY    J[\YWa
Return  :   none
Note    :
-------------------------------------------------------------------------------*/
static void MoveMouseCursor(HWND hWnd, int posX, int posY)
{
    LPTSTR      old;
    int         timer;
    RECT        rect;
    FILEMANTBL  *curFile;

    curFile = AskCurFileTbl();

    timer = NO;
    GetClientRect(hWnd, &rect);
    if(posY < 0)
    {
        SendMessage(hWnd, WM_COMMAND, MAKEWPARAM(MENU_PREVLINE, 0), 0);
        posY = 0;
        timer = YES;
    }
    else if(posY > rect.bottom)
    {
        SendMessage(hWnd, WM_COMMAND, MAKEWPARAM(MENU_NEXTLINE, 0), 0);
        posY = rect.bottom-1;
        timer = YES;
    }

    old = curFile->RgnEndPos;
    curFile->RgnEndPos = AskCursorCharPos(hWnd, posX, posY);
    if(old != curFile->RgnEndPos)
    {
        InvertTextColor(hWnd, NULL, old, curFile->RgnEndPos);
    }
    if(timer == YES)
    {
        SetTimer(hWnd, 1, 1, NULL);
    }
    return;
}


/*-------------------------------------------------------------------------------
Name    :   SetFontSize
Desc    :   tHg̑傫vZ
Param   :   none
Return  :   none
Note    :
-------------------------------------------------------------------------------*/
void SetFontSize(void)
{
    HDC     hDC;
    HFONT   hFontOld;
    SIZE    size;
    int     i;

    hDC = GetDC(hWndView);

    if(hFontttPage != NULL)
    {
        DeleteObject(hFontttPage);
    }
    hFontttPage = CreateFontIndirect(&(AskCurExtSet()->Font));

    hFontOld = NULL;
    if(hFontttPage != NULL)
    {
        hFontOld = SelectObject(hDC, hFontttPage);
    }
    GetTextExtentPoint32(hDC, _T("aAjJ,_"), 6, &size);

    size.cx /= 6;
    for(i = 0; i < ONELINE_MAXLEN; i++)
    {
        CellSep[i] = size.cx;
    }
    CharWidth = size.cx;
    CharHeight = size.cy;

    if(hFontOld != NULL)
    {
        SelectObject(hDC, hFontOld);
    }
    ReleaseDC(hWndView, hDC);

    return;
}


/*-------------------------------------------------------------------------------
Name    :   SetCellSize
Desc    :   Z̃TCYvZ
Param   :   none
Return  :   none
Note    :
-------------------------------------------------------------------------------*/
void SetCellSize(void)
{
    CharCellHeight = CharHeight + AskCurExtSet()->LineSpace;
    CharCellWidth = CharWidth;
    return;
}


/*-------------------------------------------------------------------------------
Name    :   AskCharCellHeight
Desc    :   Z̍Ԃ
Param   :   none
Return  :   
Note    :
-------------------------------------------------------------------------------*/
int AskCharCellHeight(void)
{
    return(CharCellHeight);
}


/*-------------------------------------------------------------------------------
Name    :   SetWindowSize
Desc    :   EChE̕\^svZ
Param   :   none
Return  :   none
Note    :
-------------------------------------------------------------------------------*/
void SetWindowSize(void)
{
    RECT    rect;

    GetClientRect(hWndView, &rect);
    WinLines = (rect.bottom - rect.top + CharCellHeight - 1) / CharCellHeight;
    WinColumns = (rect.right - rect.left + CharCellWidth - 1) / CharCellWidth;

    return;
}


/*-------------------------------------------------------------------------------
Name    :   RePaintView
Desc    :   WM_PAINT̏
Param   :   hWnd    EChEnh
Return  :   none
Note    :
-------------------------------------------------------------------------------*/
static void RePaintView(HWND hWnd)
{
    PAINTSTRUCT ps;
    RECT        rect;

    BeginPaint(hWnd, (LPPAINTSTRUCT) &ps);

    if(AskCurFileTbl() != NULL)
    {
        DispText(hWnd, ps.hdc, 0, WinLines);
    }
    else
    {
        GetClientRect(hWnd, &rect);
        FillRect(ps.hdc, &rect, hBrushBack);
    }

    EndPaint(hWnd, (LPPAINTSTRUCT) &ps);
    return;
}


/*-------------------------------------------------------------------------------
Name    :   ChangeDispTopLine
Desc    :   \擪sς̏
Param   :   hWnd    EChEnh
            num     ړ̕\擪sԍ
            redraw  Sĕ\tO (YES/NO)
Return  :   none
Note    :
-------------------------------------------------------------------------------*/
static void ChangeDispTopLine(HWND hWnd, int num, int redraw)
{
    int         offs;
    FILEMANTBL  *curFile;
    int         dir;
    int         i;
    int         smooth;

    curFile = AskCurFileTbl();
    offs = num - curFile->DispTopLine;
    if((redraw == YES) || (abs(offs) >= WinLines))
    {
        curFile->DispTopLine = num;
        InvalidateRgn(hWndView, NULL, TRUE);
        InvalidateRgn(GetNumWinHwnd(), NULL, TRUE);
        SetScrollBar();
    }
    else
    {
        if(((offs < 0) && (curFile->DispTopLine > 0)) ||
           ((offs > 0) && (curFile->DispTopLine < (curFile->TextLines - 1))))
        {
            if((GlobalSet.SmoothPage == NO) || (abs(offs) <= 3))
            {
                smooth = NO;
                if(abs(offs) <= 3)
                {
                    smooth = GlobalSet.SmoothLine;
                }
                curFile->DispTopLine = num;
                ScrollUpDownView(-offs, smooth);
                ScrollUpDownNum(-offs);
                SetScrollBar();
            }
            else
            {
                dir = (offs < 0) ? -1 : 1;
                for(i = 0; i < abs(offs); i++)
                {
                    curFile->DispTopLine += dir;
                    ScrollUpDownView(-dir, NO);
                }
                ScrollUpDownNum(-offs);
                SetScrollBar();
            }
        }
    }

    DispCurrentLineNum(STATUS_WIN_TOP_LINE_NUM, num);
    DispMousePos();

    return;
}


/*-------------------------------------------------------------------------------
Name    :   ChangeDispTopColumn
Desc    :   \擪ς̏
Param   :   hWnd    EChEnh
            num     ړ̕\擪sԍ
            redraw  Sĕ\tO (YES/NO)
Return  :   none
Note    :
-------------------------------------------------------------------------------*/
static void ChangeDispTopColumn(HWND hWnd, int num, int redraw)
{
    int         offs;
    FILEMANTBL  *curFile;

    curFile = AskCurFileTbl();

    offs = num - curFile->DispTopColumn;
    if((redraw == YES) || (abs(offs) >= WinColumns))
    {
        curFile->DispTopColumn = num;
        InvalidateRgn(hWndView, NULL, TRUE);
        SetScrollBar();
    }
    else
    {
        if(offs < 0)
        {
            if(curFile->DispTopColumn > 0)
            {
                curFile->DispTopColumn = num;
                ScrollLeftRightView(-offs);
                SetScrollBar();
            }
        }
        else if(offs > 0)
        {
            if(curFile->DispTopColumn < (AskCurExtSet()->FoldPos - 1))
            {
                curFile->DispTopColumn = num;
                ScrollLeftRightView(-offs);
                SetScrollBar();
            }
        }
    }
    return;
}


/*-------------------------------------------------------------------------------
Name    :   DispText
Desc    :   eLXg\
Param   :   hWnd    EChEnh
            hDC     foCXReLXg (NULL=GetDCŋ߂)
            top     \擪sԍ
            btm     \sԍ
Return  :   none
Note    :
-------------------------------------------------------------------------------*/
static void DispText(HWND hWnd, HDC hDC, int top, int btm)
{
    HDC         inDC;
    HFONT       hFontOld;
    int         posY;
    int         posX;
    LPTSTR      text;
    int         len;
    int         i;
    HDC         memDC[4];
    HBITMAP     bmp[4];
    HBITMAP     bmpOld[4];
    BITMAP      bmpx;
    RECT        rect;
    int         tab;
    LPTSTR      rgnTop;
    LPTSTR      rgnBtm;
    FILEMANTBL  *curFile;
    SIZE        size;
    int         spaceWidthPx;

    static const BYTE CrMarkBmp[2][16] = {
        {
            0xff,0x00,
            0xff,0x00,
            0xff,0x00,
            0xff,0x00,
            0xff,0x00,
            0xff,0x00,
            0xff,0x00,
            0xff,0x00
        },
        {
            0xdf,0x00,  /* __*_____ */
            0xdf,0x00,  /* __*_____ */
            0xdf,0x00,  /* __*_____ */
            0xdf,0x00,  /* __*_____ */
            0xdf,0x00,  /* __*_____ */
            0x57,0x00,  /* *_*_*___ */
            0x8f,0x00,  /* _***____ */
            0xdf,0x00   /* __*_____ */
        }
    };

    static const BYTE TabMarkBmp[2][16] = {
        {
            0xff,0x00,
            0xff,0x00,
            0xff,0x00,
            0xff,0x00,
            0xff,0x00,
            0xff,0x00,
            0xff,0x00,
            0xff,0x00
        },
        {
            0xff,0x00,  /* ________ */
            0xff,0x00,  /* ________ */
            0xff,0x00,  /* ________ */
            0x3f,0x00,  /* **______ */
            0xcf,0x00,  /* __**____ */
            0x3f,0x00,  /* **______ */
            0xff,0x00,  /* ________ */
            0xff,0x00   /* ________ */
        }
    };

    static const BYTE HanSpaceMarkBmp[1][16] = {
        {
            0xff,0x00,  /* ________ */
            0xff,0x00,  /* ________ */
            0xff,0x00,  /* ________ */
            0xff,0x00,  /* ________ */
            0xff,0x00,  /* ________ */
            0x77,0x00,  /* *_____*_ */
            0xff,0x00,  /* ________ */
            0x57,0x00   /* *_*_*_*_ */
        }
    };

    static const BYTE ZenSpaceMarkBmp[1][16] = {
        {
            0xff,0x00,  /* ________ */
            0x55,0x00,  /* *_*_*_*_ */
            0xff,0x00,  /* ________ */
            0x7d,0x00,  /* *_____*_ */
            0xff,0x00,  /* ________ */
            0x7d,0x00,  /* *_____*_ */
            0xff,0x00,  /* ________ */
            0x55,0x00   /* *_*_*_*_ */
        }
    };

//DoPrintf(_T("DispText --------------\n\n\n"));

    inDC = hDC;
    if(inDC == NULL)
    {
        hDC = GetDC(hWnd);
    }
    curFile = AskCurFileTbl();

    /* s}[N */
    memDC[0] = CreateCompatibleDC(hDC);
    bmp[0] = CreateBitmap(8, 8, 1, 1, CrMarkBmp[AskCurExtSet()->CrMark]);
    GetObject(bmp[0], sizeof(BITMAP), (LPSTR) &bmpx);
    bmpOld[0] = SelectObject(memDC[0], bmp[0]);

    /* ^u}[N */
    memDC[1] = CreateCompatibleDC(hDC);
    bmp[1] = CreateBitmap(8, 8, 1, 1, TabMarkBmp[AskCurExtSet()->TabMark]);
    GetObject(bmp[1], sizeof(BITMAP), (LPSTR) &bmpx);
    bmpOld[1] = SelectObject(memDC[1], bmp[1]);

    /* pXy[X}[N */
    memDC[2] = CreateCompatibleDC(hDC);
    bmp[2] = CreateBitmap(8, 8, 1, 1, HanSpaceMarkBmp[0]);
    GetObject(bmp[2], sizeof(BITMAP), (LPSTR) &bmpx);
    bmpOld[2] = SelectObject(memDC[2], bmp[2]);

    /* SpXy[X}[N */
    memDC[3] = CreateCompatibleDC(hDC);
    bmp[3] = CreateBitmap(16, 8, 1, 1, ZenSpaceMarkBmp[0]);
    GetObject(bmp[3], sizeof(BITMAP), (LPSTR) &bmpx);
    bmpOld[3] = SelectObject(memDC[3], bmp[3]);

    /* tHg */
    hFontOld = NULL;
    if(hFontttPage != NULL)
    {
        hFontOld = SelectObject(hDC, hFontttPage);
    }

    /* ͈͎w͈̔ */
    rgnTop = curFile->RgnStartPos;
    rgnBtm = curFile->RgnEndPos;
    CheckTopAndBtmAdrs(&rgnTop, &rgnBtm);

    /* ̑ */
    GetClientRect(hWnd, &rect);
    SetBkMode(hDC, OPAQUE);
    SetTextColor(hDC, AskCurExtSet()->TextColor);
    tab = AskCurExtSet()->TabWidth;
    CheckTopAndBtmLine(&top, &btm);

    /* 1̕߂ */
    GetTextExtentPoint32(hDC, _T(" "), 1, &size);
    spaceWidthPx = size.cx;

    /* eLXg\郋[v */
    posX = LEFT_OFFSET - (curFile->DispTopColumn * CharCellWidth);
    posY = top * CharCellHeight;
    for(i = top; i < btm; i++)
    {
        /* 炩ߏ */
        rect.top = posY;
        rect.left = 0;
        rect.right = rect.right;
        rect.bottom = posY + CharCellHeight;
        FillRect(hDC, &rect, hBrushBack);

        if((text = GetLineTop(curFile, curFile->DispTopLine + i)) != NULL)
        {
            len = GetLineLen(curFile, curFile->DispTopLine + i);
            DispOneLine(hDC, posX, posY, text, len, memDC, tab, rgnTop, rgnBtm, curFile, spaceWidthPx);
        }
        else
        {
            DispOneLine(hDC, posX, posY, _T("~"), 1, memDC, tab, rgnTop, rgnBtm, curFile, spaceWidthPx);
        }

        posY += CharCellHeight;
    }

    if(hFontOld != NULL)
    {
        SelectObject(hDC, hFontOld);
    }
    SelectObject(memDC[0], bmpOld[0]);
    SelectObject(memDC[1], bmpOld[1]);
    SelectObject(memDC[2], bmpOld[2]);
    SelectObject(memDC[3], bmpOld[3]);
    DeleteDC(memDC[0]);
    DeleteDC(memDC[1]);
    DeleteDC(memDC[2]);
    DeleteDC(memDC[3]);
    DeleteObject(bmp[0]);
    DeleteObject(bmp[1]);
    DeleteObject(bmp[2]);
    DeleteObject(bmp[3]);

    if(inDC == NULL)
    {
        ReleaseDC(hWnd, hDC);
    }
    return;
}


/*-------------------------------------------------------------------------------
Name    :   DispOneLine
Desc    :   Ps̃eLXg\
Param   :   hDC     foCXReLXg
            posX        wW
            posY        xW
            text        eLXg
            len         Ps̃eLXg̒
            memDC       s}[N
            tabWidth    ^u
            rgnTop      ͈͎w̐擪
            rgnBtm      ͈͎w̖
            curFile     t@C
            spaceWidthPx 󔒕̃sNZ
Return  :   none
Note    :
-------------------------------------------------------------------------------*/
static void DispOneLine(HDC hDC, int posX, int posY, LPTSTR text, int len, HDC memDC[], int tabWidth, LPTSTR rgnTop, LPTSTR rgnBtm, FILEMANTBL *curFile, int spaceWidthPx)
{
    LPTSTR      disp;
    int         tabs;
    int         column;
    COLORREF    backColor;
    COLORREF    textColor;
    int         tmpPosX;
    int         orgPosX;
    LPTSTR      foundPos;
    int         foundChars;
    LPTSTR      rgnPos;
    int         rgnChars;
    LPTSTR      underPos;
    int         underChars;
    int         attr;

    EXTSET  *extSet = AskCurExtSet();

//DoPrintf(_T("DispOneLine ###\n"));

    foundPos = NULL;
    foundChars = 0;
    if(RevFindStr == YES)
    {
        if((foundChars = -AskCurPosFound(curFile, text)) != 0)
        {
            foundPos = text;
        }
        else
        {
            foundPos = GetNextFoundPos(curFile, text, &foundChars);
            foundChars = -foundChars;
        }
    }

    rgnPos = NULL;
    rgnChars = 0;
    if((text < rgnBtm) && (rgnTop != rgnBtm))
    {
        if(text >= rgnTop)
        {
            rgnPos = text;
        }
        else
        {
            rgnPos = rgnTop;
        }
        rgnChars = -(rgnBtm - rgnPos);
    }

    if(AskCurPosAttr(curFile, text, &underChars, &attr) != 0)
    {
        underChars = -underChars;
        underPos = text;
    }
    else
    {
        underPos = GetNextAttrPos(curFile, text, &underChars, &attr);
        underChars = -underChars;
    }

    orgPosX = posX;
    backColor = extSet->BackColor;
    textColor = extSet->TextColor;
    disp = text;
    column = 0;
    for(; len > 0; len--)
    {
        /* ̃AhXɑ΂`FbN */
        if((text == foundPos) || (text == rgnPos) || (text == underPos))
        {
            posX = DispLineGo(hDC, posX, posY, disp, text, 0, textColor, backColor);
            disp = text;

            if(text == foundPos)
            {
                backColor = (backColor & 0xFF000000) | extSet->FindColor;
                foundChars = -foundChars;
            }
            if(text == rgnPos)
            {
                backColor |= ATTR_REGION;
                rgnChars = -rgnChars;
            }

            if(text == underPos)
            {
                backColor |= attr;
                if(backColor & ATTR_BOLD)
                {
                    textColor = extSet->HighColor;
                }
                underChars = -underChars;
            }
        }

        /* ɑ΂`FbN */
        if((*text < _T(' ')) ||         /* Rg[R[h */
           ((extSet->HanSpcMark == YES) && (*text == _T(' '))) ||
           ((extSet->ZenSpcMark == YES) && (*text == _T('@'))))
        {
            posX = DispLineGo(hDC, posX, posY, disp, text, 0, textColor, backColor);
            disp = text+1;

            if(*text == _T('\t'))           /* TAB */
            {
                tabs = GetTabWidth(tabWidth, column);
                column += tabs;
                tmpPosX = orgPosX + (column * CharCellWidth);
                while(tmpPosX < posX)
                {
                    column += tabWidth;
                    tmpPosX = orgPosX + (column * CharCellWidth);
                }
                posX = DispMarkGo(hDC, posX, posY, *text, tmpPosX, extSet->TabColor, backColor, memDC);
            }
            else if((*text == _T('\r')) || (*text == _T('\n')))
            {
                posX = DispMarkGo(hDC, posX, posY, *text, 0, extSet->CrColor, backColor, memDC);
                column++;
                if((*text == _T('\r')) && (len > 1) && (*(text+1) == _T('\n')))
                {
                    text++;
                    len--;
                    disp = text+1;
                }
            }
            else if(*text == _T(' '))
            {
                posX = DispMarkGo(hDC, posX, posY, *text, 0, extSet->HanSpcColor, backColor, memDC);
                column++;
            }
            else if(*text == _T('@'))
            {
                posX = DispMarkGo(hDC, posX, posY, *text, 0, extSet->ZenSpcColor, backColor, memDC);
                column++;
            }
            else
            {
                posX = DispMarkGo(hDC, posX, posY, *text, 0, extSet->CtrlColor, backColor, memDC);
                column++;
            }
        }
        else
        {
            column += GetCharacterWidthSimple(*text);
        }

        text++;

        if(foundChars > 0)
        {
            if(--foundChars == 0)
            {
                posX = DispLineGo(hDC, posX, posY, disp, text, 0, textColor, backColor);
                disp = text;
                backColor = (backColor & 0xFF000000) | extSet->BackColor;
                foundPos = GetNextFoundPos(curFile, text, &foundChars);
                foundChars = -foundChars;
            }
        }

        if(rgnChars > 0)
        {
            if(--rgnChars == 0)
            {
                posX = DispLineGo(hDC, posX, posY, disp, text, 0, textColor, backColor);
                disp = text;
                backColor &= ~ATTR_REGION;
            }
        }

        if(underChars > 0)
        {
            if(--underChars == 0)
            {
                posX = DispLineGo(hDC, posX, posY, disp, text, 0, textColor, backColor);
                disp = text;

                backColor &= ~attr;
                textColor = extSet->TextColor;
                underPos = GetNextAttrPos(curFile, text, &underChars, &attr);
                underChars = -underChars;
            }
        }
    }
    posX = DispLineGo(hDC, posX, posY, disp, text, 0, textColor, backColor);

    return;
}


/*-------------------------------------------------------------------------------
Name    :   GetCharacterWidthSimple
Desc    :   SppԂ
Param   :   text        
Return  :   1 or 2
Note    :
-------------------------------------------------------------------------------*/
static int GetCharacterWidthSimple(_TCHAR text)
{
    if ((text <= 0x007e) ||                         /* pL*/
        ((text >= 0xff61) && (text <= 0xff9f)))     /* pJi */
    {
        return 1;
    }
    else
    {
        return 2;
    }
}


/*-------------------------------------------------------------------------------
Name    :   DispMarkGo
Desc    :   Pŝ̎w͈͂̋L\
Param   :   hDC         foCXReLXg
            posX        wW
            posY        xW
            text        eLXg
            btmPosX     ^ȕIʒûwW(*text=='\t'̎̂)
            textColor   F
            backColor   wiF (& ATTR_REGION)
            memDC       s}[NȂ
Return  :   \̂wW
Note    :
-------------------------------------------------------------------------------*/
static int DispMarkGo(HDC hDC, int posX, int posY, _TCHAR text, int btmPosX, COLORREF textColor, COLORREF backColor, HDC memDC[])
{
    RECT        rect;
    SIZE        size;
    int         nextPosX;
    _TCHAR      tmp[2];
    HPEN        hPen;
    HPEN        hPenOld;

//DoPrintf(_T("DispMarkGo 0x%x-0x%x (%d) 0x%02x\n"), text, btm, btm-text, *text);

    GetClientRect(hWndView, &rect);
    rect.left = LEFT_OFFSET;
    SetTextColor(hDC, textColor);
    SetBkColor(hDC, (backColor & 0x00FFFFFF));

    if((text == _T('\n')) || (text == _T('\r')))
    {
        rect.left = posX;
        rect.top = posY;
        rect.right = posX + CharCellWidth;
        rect.bottom = posY + CharHeight;
        ExtTextOut(hDC, posX, posY, ETO_OPAQUE, &rect, _T(""), 0, (const int *)&CellSep);
        BitBlt(hDC, posX+((CharWidth-5)/2), posY+((CharHeight-8)/2), 5, 8, memDC[0], 0, 0, SRCCOPY);
        nextPosX = posX + CharCellWidth;
    }
    else if(text == _T('\t'))
    {
        rect.left = posX;
        rect.top = posY;
        rect.right = btmPosX;
        rect.bottom = posY + CharHeight;
        ExtTextOut(hDC, posX, posY, ETO_OPAQUE, &rect, _T(""), 0, (const int *)&CellSep);
        BitBlt(hDC, posX, posY+((CharHeight-8)/2), 5, 8, memDC[1], 0, 0, SRCCOPY);
        nextPosX = btmPosX;
    }
    else if(text == _T(' '))
    {
        rect.left = posX;
        rect.top = posY;
        rect.right = posX + CharCellWidth;
        rect.bottom = posY + CharHeight;
        ExtTextOut(hDC, posX, posY, ETO_OPAQUE, &rect, _T(""), 0, (const int *)&CellSep);
        BitBlt(hDC, posX, posY+(CharHeight-8)-1, 8, 8, memDC[2], 0, 0, SRCCOPY);
        nextPosX = posX + CharCellWidth;
    }
    else if(text == _T('@'))
    {
        rect.left = posX;
        rect.top = posY;
        rect.right = posX + (CharCellWidth*2);
        rect.bottom = posY + CharHeight;
        ExtTextOut(hDC, posX, posY, ETO_OPAQUE, &rect, _T(""), 0, (const int *)&CellSep);
        BitBlt(hDC, posX+((CharWidth*2-8)/2), posY+((CharHeight-8)/2), 8, 8, memDC[3], 0, 0, SRCCOPY);
        nextPosX = posX + (CharCellWidth*2);
    }
    else
    {
        if(text < _T(' '))
        {
            tmp[0] = _T('@') + text;
        }
        ExtTextOut(hDC, posX, posY, ETO_CLIPPED, &rect, tmp, 1, NULL);
        GetTextExtentPoint32(hDC, tmp, 1, &size);
        nextPosX = posX + size.cx;
    }

    if(backColor & ATTR_UNDER)
    {
        hPen = CreatePen(PS_SOLID, 0, textColor);
        hPenOld = (HPEN)SelectObject(hDC, hPen);
        rect.bottom = posY + CharHeight - 1;
        rect.left  = max(LEFT_OFFSET, posX);
        rect.right  = max(LEFT_OFFSET, nextPosX);
        MoveToEx(hDC, rect.left, rect.bottom, NULL);
        LineTo(hDC, rect.right, rect.bottom);
        SelectObject(hDC, hPenOld);
        DeleteObject(hPen);
    }
    if(backColor & ATTR_REGION)
    {
        rect.top = posY;
        rect.bottom = posY + CharHeight;
        rect.left  = max(LEFT_OFFSET, posX);
        rect.right  = max(LEFT_OFFSET, nextPosX);
        InvertRect(hDC, &rect);
    }
    posX = nextPosX;

    return(posX);
}


/*-------------------------------------------------------------------------------
Name    :   DispLineGo
Desc    :   Pŝ̎w͈͂̃eLXg\
Param   :   hDC         foCXReLXg
            posX        wW
            posY        xW
            text        eLXg
            btmPosX     ^ȕIʒûwW(*text=='\t'̎̂)
            textColor   F
            backColor   wiF (& ATTR_REGION)
Return  :   \̂wW
Note    :
-------------------------------------------------------------------------------*/
static int DispLineGo(HDC hDC, int posX, int posY, LPTSTR text, LPTSTR btm, int btmPosX, COLORREF textColor, COLORREF backColor)
{
    RECT    rect;
    SIZE    size;
    int     len;
    int     nextPosX;
    HPEN    hPen;
    HPEN    hPenOld;

//DoPrintf(_T("DispLineGo 0x%x-0x%x (%d) 0x%02x\n"), text, btm, btm-text, *text);

    len = btm - text;
    if(len > 0)
    {
        GetClientRect(hWndView, &rect);
        rect.left = LEFT_OFFSET;
        SetTextColor(hDC, textColor);
        SetBkColor(hDC, (backColor & 0x00FFFFFF));

        ExtTextOut(hDC, posX, posY, ETO_CLIPPED, &rect, text, len, NULL);
        GetTextExtentPoint32(hDC, text, len, &size);
        nextPosX = posX + size.cx;

        if(backColor & ATTR_UNDER)
        {
            hPen = CreatePen(PS_SOLID, 0, textColor);
            hPenOld = (HPEN)SelectObject(hDC, hPen);
            rect.bottom = posY + CharHeight - 1;
            rect.left  = max(LEFT_OFFSET, posX);
            rect.right  = max(LEFT_OFFSET, nextPosX);
            MoveToEx(hDC, rect.left, rect.bottom, NULL);
            LineTo(hDC, rect.right, rect.bottom);
            SelectObject(hDC, hPenOld);
            DeleteObject(hPen);
        }
        if(backColor & ATTR_REGION)
        {
            rect.top = posY;
            rect.bottom = posY + CharHeight;
            rect.left  = max(LEFT_OFFSET, posX);
            rect.right  = max(LEFT_OFFSET, nextPosX);
            InvertRect(hDC, &rect);
        }
        posX = nextPosX;
    }
    return(posX);
}


/*-------------------------------------------------------------------------------
Name    :   InvertTextColor
Desc    :   w͈͂̃eLXg𔽓]\
Param   :   hWnd        EChEnh
            hDC         foCXReLXg (NULL=GetDCŋ߂)
            rgnTop      ͈͎w̐擪
            rgnBtm      ͈͎w̖
Return  :   none
Note    :
-------------------------------------------------------------------------------*/
static void InvertTextColor(HWND hWnd, HDC hDC, LPTSTR rgnTop, LPTSTR rgnBtm)
{
    LPTSTR      text;
    LPTSTR      lineTop;
    int         len;
    int         i;
    RECT        reverseRect;
    HDC         inDC;
    int         column;
    int         tabWidth;
    FILEMANTBL  *curFile;
    HFONT       hFontOld;
    int         posX;
    int         tmpPosX;
    LPTSTR      chkTop;

    CheckTopAndBtmAdrs(&rgnTop, &rgnBtm);
    curFile = AskCurFileTbl();
    tabWidth = AskCurExtSet()->TabWidth;

    inDC = hDC;
    if(inDC == NULL)
    {
        hDC = GetDC(hWnd);
    }

    hFontOld = NULL;
    if(hFontttPage != NULL)
    {
        hFontOld = SelectObject(hDC, hFontttPage);
    }

    for(i = 0; i < WinLines; i++)
    {
        if((lineTop = GetLineTop(curFile, curFile->DispTopLine + i)) != NULL)
        {
            len = GetLineLen(curFile, curFile->DispTopLine + i);

            text = lineTop;
            reverseRect.left = -1;
            reverseRect.right = -1;
            column = 0;
            posX = 0;
            chkTop = text;
            for(; len > 0; len--)
            {
                if((reverseRect.left == -1) && (text >= rgnTop) && (text < rgnBtm))
                {
                    reverseRect.left = posX + GetTextPixelWidth(hDC, chkTop, text-chkTop);
                }
                else if((reverseRect.left != -1) && (reverseRect.right == -1) && (text >= rgnBtm))
                {
                    reverseRect.right = posX + GetTextPixelWidth(hDC, chkTop, text-chkTop);
                    break;
                }

                if(*text == _T('\t'))
                {
                    posX += GetTextPixelWidth(hDC, chkTop, text-chkTop);

                    column += GetTabWidth(tabWidth, column);
                    tmpPosX = column * CharCellWidth;
                    while(tmpPosX < posX)
                    {
                        column += tabWidth;
                        tmpPosX = (column * CharCellWidth);
                    }
                    posX = tmpPosX;
                    text++;
                    chkTop = text;
                }
                else if(*text == _T('\r'))
                {
                    posX += GetTextPixelWidth(hDC, chkTop, text-chkTop);
                    posX += CharCellWidth;
                    text++;
                    column++;
                    chkTop = text;
                    if((len > 1) && (*text == _T('\n')))
                    {
                        text++;
                        len--;
                        chkTop = text;
                    }
                }
                else if(*text == _T('\n'))
                {
                    posX += GetTextPixelWidth(hDC, chkTop, text-chkTop);
                    posX += CharCellWidth;
                    text++;
                    column++;
                    chkTop = text;
                }
                else
                {
                    column += GetCharacterWidthSimple(*text);
                    text++;
                }
            }

            if(reverseRect.left != -1)
            {
                if(reverseRect.right == -1)
                {
                    reverseRect.right = posX + GetTextPixelWidth(hDC, chkTop, text-chkTop);
                }
                reverseRect.top = CharCellHeight * i;
                reverseRect.bottom = reverseRect.top + CharHeight;
                reverseRect.left  = max(LEFT_OFFSET, reverseRect.left + LEFT_OFFSET - (curFile->DispTopColumn * CharCellWidth));
                reverseRect.right  = max(LEFT_OFFSET, reverseRect.right + LEFT_OFFSET - (curFile->DispTopColumn * CharCellWidth));
                InvertRect(hDC, &reverseRect);
            }
        }
    }

    if(hFontOld != NULL)
    {
        SelectObject(hDC, hFontOld);
    }
    if(inDC == NULL)
    {
        ReleaseDC(hWnd, hDC);
    }
    return;
}


/*-------------------------------------------------------------------------------
Name    :   GetTextPixelWidth
Desc    :   w͈͂̃eLXg̃sNZԂ
Param   :   hDC     foCXReLXg
            text    eLXg
            len     eLXg̕
Return  :   sNZ
Note    :
-------------------------------------------------------------------------------*/
static int GetTextPixelWidth(HDC hDC, LPTSTR text, int len)
{
    SIZE    size;

    if ((len == 1) && (*text <= 0x001f))
    {
        size.cx = CharWidth;
    }
    else
    {
        GetTextExtentPoint32(hDC, text, len, &size);
    }
    return(size.cx);
}


/*-------------------------------------------------------------------------------
Name    :   CheckTopAndBtmAdrs
Desc    :   AhX̑O֌W𒲐
Param   :   top     擪
            btm     
Return  :   none
Note    :
-------------------------------------------------------------------------------*/
static void CheckTopAndBtmAdrs(LPTSTR *top, LPTSTR *btm)
{
    LPTSTR  tmp;

    if(*btm < *top)
    {
        tmp = *top;
        *top = *btm;
        *btm = tmp;
    }
    return;
}


/*-------------------------------------------------------------------------------
Name    :   CheckTopAndBtmLine
Desc    :   sԍ̑O֌W𒲐
Param   :   top     擪
            btm     
Return  :   none
Note    :
-------------------------------------------------------------------------------*/
static void CheckTopAndBtmLine(int *top, int *btm)
{
    int     tmp;

    if(*btm < *top)
    {
        tmp = *top;
        *top = *btm;
        *btm = tmp;
    }
    return;
}


/*-------------------------------------------------------------------------------
Name    :   AskRegSelected
Desc    :   ͈͎w肪sĂ邩ǂԂ
Param   :   none
Return  :   Xe[^X (YES/NO)
Note    :
-------------------------------------------------------------------------------*/
int AskRegSelected(void)
{
    int     sts;

    sts = NO;
    if(AskCurFileTbl()->RgnStartPos != NULL)
    {
        sts = YES;
    }
    return(sts);
}


/*-------------------------------------------------------------------------------
Name    :   ScrollUpDownView
Desc    :   ㉺XN[
Param   :   num     XN[s (smooth=1̏ꍇ1ŝ)
            smooth  X[XXN[ǂ (YES/NO)
Return  :   none
Note    :
-------------------------------------------------------------------------------*/
static void ScrollUpDownView(int num, int smooth)
{
    RECT    rect;
    RECT    upRect;
    HDC     hDC;
    int     dir;
    int     step;
    int     lines;
    int     linesX;
    int     curX;
    int     cur;
    int     i;

    hDC = GetDC(hWndView);
    GetClientRect(hWndView, &rect);
    rect.left += LEFT_OFFSET;

    if(smooth == NO)
    {
        ScrollDC(hDC, 0, num * CharCellHeight, &rect, &rect, NULL, &upRect);
        RedrawWindow(hWndView, &upRect, NULL, RDW_UPDATENOW | RDW_INVALIDATE | RDW_NOINTERNALPAINT);
    }
    else
    {
        dir = (num < 0) ? -1 : 1;
        lines = abs(num) * CharCellHeight;
        linesX = lines * 1000;
        step = linesX / GlobalSet.SmoothRate;
        cur = 0;
        curX = 0;

        for(i = 0; i < GlobalSet.SmoothRate; i++)
        {
            curX += step;
            if(cur < (curX/1000))
            {
                ScrollDC(hDC, 0, (curX/1000 - cur) * dir, &rect, &rect, NULL, &upRect);
                FillRect(hDC, &upRect, hBrushBack);
                cur = curX/1000;
            }
        }

        if(cur < lines)
        {
            ScrollDC(hDC, 0, (lines - cur) * dir, &rect, &rect, NULL, &upRect);
            FillRect(hDC, &upRect, hBrushBack);
        }

        if(dir == -1)
        {
            rect.top = rect.bottom - lines;
        }
        else
        {
            rect.bottom = rect.top + lines;
        }
        RedrawWindow(hWndView, &rect, NULL, RDW_UPDATENOW | RDW_INVALIDATE | RDW_NOINTERNALPAINT);
    }

    ReleaseDC(hWndView, hDC);

    return;
}


/*-------------------------------------------------------------------------------
Name    :   ScrollLeftRightView
Desc    :   EXN[
Param   :   num     XN[s
Return  :   none
Note    :
-------------------------------------------------------------------------------*/
static void ScrollLeftRightView(int num)
{
    RECT    rect;
    RECT    upRect;
    HDC     hDC;

    hDC = GetDC(hWndView);
    GetClientRect(hWndView, &rect);
    rect.left += LEFT_OFFSET;
    ScrollDC(hDC, num * CharCellWidth, 0, &rect, &rect, NULL, &upRect);
    ReleaseDC(hWndView, hDC);

    RedrawWindow(hWndView, &upRect, NULL, RDW_UPDATENOW | RDW_INVALIDATE | RDW_NOINTERNALPAINT);

    return;
}


/*-------------------------------------------------------------------------------
Name    :   SetScrollBar
Desc    :   XN[o[̃Zbg
Param   :   none
Return  :   none
Note    :
-------------------------------------------------------------------------------*/
void SetScrollBar(void)
{
    SCROLLINFO  scrInfo;
    FILEMANTBL  *curFile;

    curFile = AskCurFileTbl();

    /* XN[o[ */
    scrInfo.cbSize = sizeof(SCROLLINFO);
    scrInfo.fMask = SIF_RANGE | SIF_PAGE | SIF_POS | SIF_DISABLENOSCROLL;
    scrInfo.nMax = curFile->TextLines / curFile->SbarStep;
    scrInfo.nMin = 0;
    scrInfo.nPage = WinLines / curFile->SbarStep;
    scrInfo.nPos = curFile->DispTopLine / curFile->SbarStep;
    SetScrollInfo(hWndView, SB_VERT, &scrInfo, TRUE);

    /* XN[o[ */
    scrInfo.cbSize = sizeof(SCROLLINFO);
    scrInfo.fMask = SIF_RANGE | SIF_PAGE | SIF_POS  | SIF_DISABLENOSCROLL;
    scrInfo.nMin = 0;
    if(AskCurExtSet()->FoldRight == NO)
    {
        scrInfo.nMax = AskCurExtSet()->FoldPos;
        scrInfo.nPage = WinColumns;
        scrInfo.nPos = curFile->DispTopColumn;
    }
    else
    {
        scrInfo.nMax = 0;
        scrInfo.nPage = 0;
        scrInfo.nPos = 0;
    }
    SetScrollInfo(hWndView, SB_HORZ, &scrInfo, TRUE);

    return;
}


/*-------------------------------------------------------------------------------
Name    :   SetDispTopWindowMatch
Desc    :   EChẼTCYɂ悤ɕ\擪ʒu𒲐߂
Param   :   none
Return  :   none
Note    :
-------------------------------------------------------------------------------*/
void SetDispTopWindowMatch(void)
{
    SCROLLINFO  scrInfo;
    EXTSET      *curExt;
    FILEMANTBL  *curFile;

    curExt = AskCurExtSet();
    curFile = AskCurFileTbl();

    scrInfo.cbSize = sizeof(SCROLLINFO);
    scrInfo.fMask = SIF_POS;
    GetScrollInfo(hWndView, SB_HORZ, &scrInfo);

    if(scrInfo.nPos >= curExt->FoldPos - 1)
    {
        scrInfo.nPos = curExt->FoldPos - 1;
    }
    ChangeDispTopColumn(hWndView, scrInfo.nPos, NO);

    scrInfo.cbSize = sizeof(SCROLLINFO);
    scrInfo.fMask = SIF_POS;
    GetScrollInfo(hWndView, SB_VERT, &scrInfo);

    if(scrInfo.nPos >= curFile->TextLines - 1)
    {
        scrInfo.nPos = curFile->TextLines - 1;
    }
    ChangeDispTopLine(hWndView, scrInfo.nPos, NO);

    return;
}


/*-------------------------------------------------------------------------------
Name    :   AskCursorCharPos
Desc    :   J[\ʒu當ʒu߂
Param   :   hWnd    EChEnh
            posX    XW
            posY    YW
Return  :   ʒu
Note    :
-------------------------------------------------------------------------------*/
static LPTSTR AskCursorCharPos(HWND hWnd, int posX, int posY)
{
    HDC         hDC;
    HFONT       hFontOld;
    int         tabWidth;
    LPTSTR      lineTop;
    LPTSTR      text;
    int         len;
    int         column;
    FILEMANTBL  *curFile;
    int         add;
    int         chkPosX;
    int         tmpPosX;

    curFile = AskCurFileTbl();
    tabWidth = AskCurExtSet()->TabWidth;

    hDC = GetDC(hWnd);
    hFontOld = NULL;
    if(hFontttPage != NULL)
    {
        hFontOld = SelectObject(hDC, hFontttPage);
    }

    posY /= CharCellHeight;
    posX -= LEFT_OFFSET - (curFile->DispTopColumn * CharCellWidth);

    if((lineTop = GetLineTop(curFile, curFile->DispTopLine + posY)) != NULL)
    {
        len = GetLineLen(curFile, curFile->DispTopLine + posY);
        text = lineTop;
        column = 0;
        chkPosX = 0;
        for(; len > 0; )
        {
            if(*text == _T('\t'))       /* TAB */
            {
                column += GetTabWidth(tabWidth, column);
                tmpPosX = column * CharCellWidth;
                while(tmpPosX < chkPosX)
                {
                    column += tabWidth;
                    tmpPosX = (column * CharCellWidth);
                }
                chkPosX = tmpPosX;
            }
            else
            {
                chkPosX += GetTextPixelWidth(hDC, text, 1);
                column += GetCharacterWidthSimple(*text);

            }
            add = 1;

            if(chkPosX >= posX)
            {
                break;
            }
            text += add;
            len -= add;
        }
    }
    else
    {
        text = curFile->TextData + curFile->TextSize;   /*  */
    }

    if(hFontOld != NULL)
    {
        SelectObject(hDC, hFontOld);
    }
    ReleaseDC(hWnd, hDC);

    return(text);
}


/*-------------------------------------------------------------------------------
Name    :   CopyToClipBoard
Desc    :   Nbv{[hփRs[
Param   :   hWnd        EChEnh
            rgnTop      ͈͎w̐擪
            rgnBtm      ͈͎w̖
            quote       p
Return  :   none
Note    :
-------------------------------------------------------------------------------*/
static void CopyToClipBoard(HWND hWnd, LPTSTR rgnTop, LPTSTR rgnBtm, LPTSTR quote)
{
    int     len;
    void    *gBuf;
    HGLOBAL hGlobal;

    CheckTopAndBtmAdrs(&rgnTop, &rgnBtm);

    if(OpenClipboard(hWnd))
    {
        if(EmptyClipboard())
        {
            len = GetStrLenOfCopyToClipboard(rgnTop, (rgnBtm-rgnTop), quote);
            if((hGlobal = GlobalAlloc(GMEM_DDESHARE | GMEM_MOVEABLE | GMEM_ZEROINIT, (len+1) * sizeof(_TCHAR))) != NULL)
            {
                if((gBuf = GlobalLock(hGlobal)) != NULL)
                {
                    CopyStrToClipboard(gBuf, rgnTop, rgnBtm-rgnTop, quote);
                    *((LPTSTR )gBuf + len) = NUL;

                    GlobalUnlock(hGlobal);
                    SetClipboardData(CF_UNICODETEXT, hGlobal);
                }
            }
        }
        CloseClipboard();
    }
    return;
}


/*-------------------------------------------------------------------------------
Name    :   AppendToClipBoard
Desc    :   Nbv{[h֒ǉRs[
Param   :   hWnd        EChEnh
            rgnTop      ͈͎w̐擪
            rgnBtm      ͈͎w̖
Return  :   none
Note    :
-------------------------------------------------------------------------------*/
static void AppendToClipBoard(HWND hWnd, LPTSTR rgnTop, LPTSTR rgnBtm)
{
    int     len;
    int     size;
    LPTSTR  gBuf;
    HGLOBAL hGlobal;
    LPTSTR  tmp;

    CheckTopAndBtmAdrs(&rgnTop, &rgnBtm);

    if(OpenClipboard(hWnd))
    {
        if((hGlobal = GetClipboardData(CF_UNICODETEXT)) != NULL)
        {
            if((gBuf = GlobalLock(hGlobal)) != NULL)
            {
                size = _tcslen(gBuf);
                len = GetStrLenOfCopyToClipboard(rgnTop, (rgnBtm-rgnTop), _T(""));

                /* GlobalReAlloc()găNbN{[h̃g  */
                /* Ă܂Ȃ̂ŁAʂ̃obt@ɕۑĂ */
                /* ăNbv{[hȂ                         */
                if((tmp = Malloc((size + 1) * sizeof(_TCHAR))) != NULL)
                {
                    _tcscpy(tmp, gBuf);

                    GlobalUnlock(hGlobal);
                    if(EmptyClipboard())
                    {
                        if((hGlobal = GlobalAlloc(GMEM_DDESHARE | GMEM_MOVEABLE | GMEM_ZEROINIT, (size + len + 1) * sizeof(_TCHAR))) != NULL)
                        {
                            if((gBuf = GlobalLock(hGlobal)) != NULL)
                            {
                                _tcscpy(gBuf, tmp);
                                CopyStrToClipboard(gBuf+size, rgnTop, rgnBtm-rgnTop, _T(""));
                                *(gBuf+size+len) = NUL;

                                GlobalUnlock(hGlobal);
                                SetClipboardData(CF_UNICODETEXT , hGlobal);
                            }
                        }
                    }
                    free(tmp);
                }
            }
        }
        CloseClipboard();
    }
    return;
}


/*-------------------------------------------------------------------------------
Name    :   CopyToBuffer
Desc    :   w̃obt@փRs[
Param   :   hWnd        EChEnh
            rgnTop      ͈͎w̐擪
            rgnBtm      ͈͎w̖
            buf         [out] f[^
            max         obt@̃TCY
Return  :   none
Note    :
-------------------------------------------------------------------------------*/
static void CopyToBuffer(HWND hWnd, LPTSTR rgnTop, LPTSTR rgnBtm, LPTSTR buf, int max)
{
    memset(buf, NUL, max * sizeof(_TCHAR));
    max--;

    CheckTopAndBtmAdrs(&rgnTop, &rgnBtm);
    if((rgnBtm - rgnTop) < max)
    {
        max = (rgnBtm - rgnTop) * sizeof(_TCHAR);
    }
    else
    {
        max *= sizeof(_TCHAR);
    }
    memcpy(buf, rgnTop, max);

    return;
}


/*-------------------------------------------------------------------------------
Name    :   AskDispFindRev
Desc    :   ʂ̔]\sȂǂԂ
Param   :   none
Return  :   Xe[^X (YES/NO)
Note    :
-------------------------------------------------------------------------------*/
int AskDispFindRev(void)
{
    return(RevFindStr);
}


/*-------------------------------------------------------------------------------
Name    :   FindDialogCallBack
Desc    :   _CAÕR[obN
Param   :   hDlg    EChEnh
            message bZ[Wԍ
            wParam  bZ[W WPARAM 
            lParam  bZ[W LPARAM 
Return  :   Xe[^X (TRUE/FALSE)
Note    :
-------------------------------------------------------------------------------*/
static BOOL CALLBACK FindDialogCallBack(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
    static FINDDLGDATA  *find;
    LPTSTR              pos;

    switch (message)
    {
        case WM_INITDIALOG :
            find = (FINDDLGDATA *)lParam;
            SendDlgItemMessage(hDlg, FIND_EDIT, CB_LIMITTEXT, FINDSTR_LEN, 0);
            pos = GlobalSet.FindHist;
            while(*pos != NUL)
            {
                SendDlgItemMessage(hDlg, FIND_EDIT, CB_ADDSTRING, 0, (LPARAM)pos);
                pos += _tcslen(pos) + 1;
            }
            SendDlgItemMessage(hDlg, FIND_EDIT, CB_SETCURSEL, 0, 0);
            if(_tcslen(find->Str) > 0)
            {
                SendDlgItemMessage(hDlg, FIND_EDIT, WM_SETTEXT, 0, (LPARAM)find->Str);
            }
            SendDlgItemMessage(hDlg, FIND_SELCASE, BM_SETCHECK, find->CapsFlg, 0);
            SendDlgItemMessage(hDlg, FIND_REGEXP, BM_SETCHECK, find->Regexp, 0);
            if(find->Regexp == YES)
            {
                EnableWindow(GetDlgItem(hDlg, FIND_SELCASE), FALSE);
            }
            return(TRUE);

        case WM_COMMAND :
            switch(GET_WM_COMMAND_ID(wParam, lParam))
            {
                case IDOK :
                    SendDlgItemMessage(hDlg, FIND_EDIT, WM_GETTEXT, FINDSTR_LEN+1, (LPARAM)find->Str);
                    find->CapsFlg = SendDlgItemMessage(hDlg, FIND_SELCASE, BM_GETCHECK, 0, 0);
                    find->Regexp = SendDlgItemMessage(hDlg, FIND_REGEXP, BM_GETCHECK, 0, 0);
                    EndDialog(hDlg, YES);
                    break;

                case IDCANCEL :
                    EndDialog(hDlg, NO);
                    break;

                case FIND_REGEXP :
                    if(SendDlgItemMessage(hDlg, FIND_REGEXP, BM_GETCHECK, 0, 0) == 1)
                    {
                        EnableWindow(GetDlgItem(hDlg, FIND_SELCASE), FALSE);
                    }
                    else
                    {
                        EnableWindow(GetDlgItem(hDlg, FIND_SELCASE), TRUE);
                    }
                    break;

                case IDHELP :
                    hWndHelp = HtmlHelp(NULL, AskHelpFilePath(), HH_HELP_CONTEXT, IDH_HELP_TOPIC_0000006);
                    break;

            }
            return(TRUE);
    }
    return(FALSE);
}


/*-------------------------------------------------------------------------------
Name    :   SetStrToFindHist
Desc    :   qXgɕo^
Param   :   str     
Return  :   none
Note    :
-------------------------------------------------------------------------------*/
static void SetStrToFindHist(LPTSTR str)
{
    DeleteStrFromMultiStr(str, GlobalSet.FindHist);
    while(CountMultiStrNum(GlobalSet.FindHist) >= FINDHIST_NUM)
    {
        DeleteNumFromMultiStr(FINDHIST_NUM-1, GlobalSet.FindHist);
    }
    InsertStrToMultiStr(str, GlobalSet.FindHist, FINDHIST_NUM*(FINDSTR_LEN+1)+1);
    return;
}


/*-------------------------------------------------------------------------------
Name    :   DispMouseCaret
Desc    :   }EXJ[\̃Jbg\
Param   :   hWnd        EChEnh
            posX        }EXJ[\Xʒu
            posY        }EXJ[\Yʒu
Return  :   none
Note    :
-------------------------------------------------------------------------------*/
static void DispMouseCaret(HWND hWnd, int posX, int posY)
{
    HDC         hDC;
    LPTSTR      text;
    LPTSTR      lineTop;
    int         len;
    RECT        reverseRect;
    int         column;
    int         tabWidth;
    FILEMANTBL  *curFile;
    HFONT       hFontOld;
    LPTSTR      rgnTop;
    int         add;
    int         chkPosX;
    int         nextPosX;

    DispMousePos();

    if(GlobalSet.DispCaret == YES)
    {
        hDC = GetDC(hWnd);
        curFile = AskCurFileTbl();
        tabWidth = AskCurExtSet()->TabWidth;

        hFontOld = NULL;
        if(hFontttPage != NULL)
        {
            hFontOld = SelectObject(hDC, hFontttPage);
        }

        posY /= CharCellHeight;
        posX -= LEFT_OFFSET - (curFile->DispTopColumn * CharCellWidth);

        /* }EXJ[\ʒu̕߂ */
        reverseRect.left = -1;
        reverseRect.right = -1;
        rgnTop = NULL;
        if((lineTop = GetLineTop(curFile, curFile->DispTopLine + posY)) != NULL)
        {
            len = GetLineLen(curFile, curFile->DispTopLine + posY);
            text = lineTop;
            column = 0;
            chkPosX = 0;
            for(; len > 0; )
            {
                if(*text == _T('\t'))       /* TAB */
                {
                    column += GetTabWidth(tabWidth, column);
                    nextPosX = column * CharCellWidth;
                    while(nextPosX < chkPosX)
                    {
                        column += tabWidth;
                        nextPosX = (column * CharCellWidth);
                    }
                }
                else
                {
                    nextPosX = chkPosX + GetTextPixelWidth(hDC, text, 1);
                    column++;
                }
                add = 1;

                if((posX >= chkPosX) && (posX <= nextPosX))
                {
                    reverseRect.left = chkPosX;
                    reverseRect.right = nextPosX;
                    rgnTop = text;
                    break;
                }

                chkPosX = nextPosX;
                text += add;
                len -= add;
            }
        }

        if(CaretPos != rgnTop)
        {
            if(CaretPos != NULL)
            {
                /*  */
                InvertRect(hDC, &CaretRect);
                CaretPos = NULL;
            }
            if(reverseRect.left != -1)
            {
                /* \ */
                reverseRect.top = CharCellHeight * posY;
                reverseRect.bottom = reverseRect.top + CharHeight;
                reverseRect.left  = max(LEFT_OFFSET, reverseRect.left + LEFT_OFFSET - (curFile->DispTopColumn * CharCellWidth));
                reverseRect.right  = max(LEFT_OFFSET, reverseRect.right + LEFT_OFFSET - (curFile->DispTopColumn * CharCellWidth));
                InvertRect(hDC, &reverseRect);

                CaretRect.top = reverseRect.top;
                CaretRect.bottom = reverseRect.bottom;
                CaretRect.left = reverseRect.left;
                CaretRect.right = reverseRect.right;
                CaretPos = rgnTop;
            }
        }

        if(hFontOld != NULL)
        {
            SelectObject(hDC, hFontOld);
        }
        ReleaseDC(hWnd, hDC);
    }
    return;
}


/*-------------------------------------------------------------------------------
Name    :   EraseMouseCaret
Desc    :   }EXJ[\̃Jbg
Param   :   hWnd        EChEnh
Return  :   none
Note    :
-------------------------------------------------------------------------------*/
static void EraseMouseCaret(HWND hWnd)
{
    HDC     hDC;

    if(CaretPos != NULL)
    {
        hDC = GetDC(hWnd);
        InvertRect(hDC, &CaretRect);
        CaretPos = NULL;
        ReleaseDC(hWnd, hDC);
    }
    return;
}


/*-------------------------------------------------------------------------------
Name    :   DispMousePos
Desc    :   }EXJ[\̈ʒu\
Param   :   none
Return  :   none
Note    :
-------------------------------------------------------------------------------*/
void DispMousePos(void)
{
    FILEMANTBL  *curFile;
    POINT       pos;
    RECT        rect;

    GetClientRect(hWndView, &rect);
    GetCursorPos(&pos);
    ScreenToClient(hWndView, &pos);
    if(((pos.y >= 0) && (pos.y < rect.bottom)) &&
       ((pos.x >= 0) && (pos.x < rect.right)))
    {
        curFile = AskCurFileTbl();
        pos.y /= CharCellHeight;
        DispCurrentLineNum(STATUS_WIN_MOUSE_LINE_NUM, curFile->DispTopLine + pos.y);
    }
    return;
}


